/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#include <cassert>
#include <memory>

#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/data/Data.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/utils/Types.h"
#include "aidge/utils/ErrorHandling.hpp"


void Aidge::OperatorTensor::associateInput(const Aidge::IOIndex_t inputIdx, const std::shared_ptr<Aidge::Data>& data) {
    if (inputIdx >= nbInputs()) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "%s Operator has %hu inputs", type().c_str(), nbInputs());
    }
    if (strcmp((data)->type(), Tensor::Type) != 0) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "Input data must be of Tensor type");
    }
    mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
}

void Aidge::OperatorTensor::setInput(const Aidge::IOIndex_t inputIdx, const std::shared_ptr<Aidge::Data>& data) {
    if (strcmp(data->type(), "Tensor") != 0) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "%s Operator only accepts Tensors as inputs", type().c_str());
    }
    if (getInput(inputIdx)) {
        *mInputs[inputIdx] = *std::dynamic_pointer_cast<Tensor>(data);
    } else {
        mInputs[inputIdx] = std::make_shared<Tensor>(*std::dynamic_pointer_cast<Tensor>(data));
    }
}

Aidge::OperatorTensor::~OperatorTensor() = default;

void Aidge::OperatorTensor::setInput(const Aidge::IOIndex_t inputIdx, std::shared_ptr<Aidge::Data>&& data) {
    if (strcmp(data->type(), "Tensor") != 0) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "%s Operator only accepts Tensors as inputs", type().c_str());
    }
    if (getInput(inputIdx)) {
        *mInputs[inputIdx] = std::move(*std::dynamic_pointer_cast<Tensor>(data));
    } else {
        mInputs[inputIdx] = std::make_shared<Tensor>(std::move(*std::dynamic_pointer_cast<Tensor>(data)));
    }
}

const std::shared_ptr<Aidge::Tensor>& Aidge::OperatorTensor::getInput(const Aidge::IOIndex_t inputIdx) const {
    if (inputIdx >= nbInputs()) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "%s Operator has %hu inputs", type().c_str(), nbInputs());
    }
    return mInputs[inputIdx];
}

void Aidge::OperatorTensor::setOutput(const Aidge::IOIndex_t outputIdx, const std::shared_ptr<Aidge::Data>& data) {
    if (strcmp(data->type(), "Tensor") != 0) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "%s Operator only accepts Tensors as inputs", type().c_str());
    }
    if (outputIdx >= nbOutputs()) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "%s Operator has %hu outputs", type().c_str(), nbOutputs());
    }
    *mOutputs[outputIdx] = *std::dynamic_pointer_cast<Tensor>(data);
}

void Aidge::OperatorTensor::setOutput(const Aidge::IOIndex_t outputIdx, std::shared_ptr<Aidge::Data>&& data) {
    if (strcmp(data->type(), "Tensor") != 0) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "%s Operator only accepts Tensors as inputs", type().c_str());
    }
    if (outputIdx >= nbOutputs()) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "%s Operator has %hu outputs", type().c_str(), nbOutputs());
    }
    *mOutputs[outputIdx] = std::move(*std::dynamic_pointer_cast<Tensor>(data));
}

const std::shared_ptr<Aidge::Tensor>& Aidge::OperatorTensor::getOutput(const Aidge::IOIndex_t outputIdx) const {
    if (outputIdx >= nbOutputs()) {
        AIDGE_THROW_OR_ABORT(std::runtime_error, "%s Operator has %hu outputs", type().c_str(), nbOutputs());
    }
    return mOutputs[outputIdx];
}


void Aidge::OperatorTensor::computeOutputDims() {
    // check inputs have been associated
    bool associated = (nbInputs() > 0); // do not compute anything if no input
    for (IOIndex_t i = 0; i < nbInputs(); ++i) {
        if (!getInput(i)) {
            AIDGE_THROW_OR_ABORT(std::runtime_error, "Every input should be associated with a Tensor");
        }
        associated &= !(getInput(i)->empty());
    }
    if (associated) {
        const auto expectedDims =  getInput(0)->dims();
        for (std::size_t i = 1; i < nbInputs(); ++i) {
            if (expectedDims != getInput(i)->dims()) {
                AIDGE_THROW_OR_ABORT(std::runtime_error, "Operator's inputs should have the same dimensions");
            }
        }
        mOutputs[0]->resize(expectedDims);
    }
}

bool Aidge::OperatorTensor::outputDimsForwarded() const {
    bool forwarded = true;
    // check both inputs and outputs have been filled
    for (IOIndex_t i = 0; i < nbInputs(); ++i) {
        forwarded &= mInputs[i] ? !(getInput(i)->empty()) : false;
    }
    for (IOIndex_t i = 0; i < nbOutputs(); ++i) {
        forwarded &= !(getOutput(i)->empty());
    }
    return forwarded;
}

void Aidge::OperatorTensor::setDataType(const DataType& dataType) const {
    for (IOIndex_t i = 0; i < nbOutputs(); ++i) {
        getOutput(i)->setDataType(dataType);
    }
    for (IOIndex_t i = 0; i < nbInputs(); ++i) {
        if (!getInput(i)) {
            AIDGE_THROW_OR_ABORT(std::runtime_error, "Input was not set");
        }
        else {
            getInput(i)->setDataType(dataType);
        }
    }
}